/*
 * Copyright (c) 2004 Paul de Weerd <weerd@weirdnet.nl>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <machine/sysarch.h>

/* function prototypes */
static void usage ( char * );
static void hci ( void );
void hcicall ( void );
static void error ( int );
static void fan ( char * );
static void backlight ( char * );
static void power ( void );
static void suspend ( char * );
static void display ( void );
static void help ( void );
static void selectbay ( char * );
static void sysfifo ( char * );
static void panel ( void );
static void irda ( char * );
static void video ( char * );
static void lock ( char * ); 

/* variables to hold the registers */
u_long reg_ax;
u_long reg_bx;
u_long reg_cx;

static void
usage ( char* progname )
{
	fprintf( stderr, "usage: %s <device> [action]\n", progname );
	fprintf( stderr, "    or %s help\n", progname );
	exit ( 1 );
}

static void
help ()
{
	printf ( "tosh <device> [action]\n" );
	printf ( "device 'backlight' has actions 'status', 'on', 'off'\n" );
	printf ( "device 'power' has no actions\n" );
	printf ( "device 'fan' has actions 'status', 'on', 'off'\n" );
	printf ( "device 'suspend' has actions 'status', 'on'\n" );
	printf ( "device 'display' has no actions\n" );
	printf ( "device 'selectbay' has actions 'internal', 'dock', 'dockbay'\n" );
	printf ( "device 'sysfifo' has actions 'status', 'on', 'off'\n" );
	printf ( "device 'panel' has no actions\n" );
	printf ( "device 'irda' has actions 'status', 'sir', 'fir'\n" );
	printf ( "device 'video' has actions 'status', 'internal', 'external', 'simultaneous'\n" );
	printf ( "device 'lock' has actions 'builtin', 'selectbay', 'dock', 'dockbay'\n" );
}

static void
hci ( void )
{
	if ( 0 != i386_iopl ( 3 ))
	{
		fprintf (stderr, "tosh: hci: This program requires root privileges.\n" );
		exit ( 2 );
	} 

	hcicall ();

	i386_iopl ( 0 );
}

static void
error ( int errorcode )
{
	switch ( errorcode ) {
		case 0x80:
			fprintf ( stderr, "System does not support this function\n" );
			break;
		case 0x83:
			fprintf ( stderr, "Input data error\n" );
			break;
		case 0x84:
			fprintf ( stderr, "Write protected\n" );
			break;
		case 0x8c:
			fprintf ( stderr, "System event FIFO empty\n" );
			break;
		default:
			fprintf ( stderr, "Unknown error\n" );
			break;
	}
	exit ( errorcode );
}
		
static void
backlight ( char *instruction )
{
	reg_bx = 0x0002;

	if ( 0 == strcmp ( instruction, "status" )) {
		reg_ax = 0xfe00;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			if ( 0 == ( reg_cx & 0xffff )) {
				printf ( "Backlight is off\n" );
			} else {
				printf ( "Backlight is on\n" );
			};
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "on" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0001;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Backlight turned on\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "off" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0000;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Backlight turned off\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else {
		printf ( "No such action\n" );
		exit ( 1 );
	};
}

static void
power ()
{
	reg_bx = 0x0003;

	reg_ax = 0xfe00;
	hci ();
	if ( 0 == ( reg_ax & 0xff00 ) ) {
		if ( 4 == ( reg_cx & 0xffff )) {
			printf ( "AC adapter online\n" );
		} else {
			printf ( "AC adapter offline\n" );
		};
	} else {
		error ( ( reg_ax & 0xff00 ) / 256 );
	};
}

static void
fan ( char *instruction )
{
	reg_bx = 0x0004;

	if ( 0 == strcmp ( instruction, "status" )) {
		reg_ax = 0xfe00;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			if ( 0 == ( reg_cx & 0xffff )) {
				printf ( "Fan is off\n" );
			} else {
				printf ( "Fan is on\n" );
			};
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "on" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0001;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Fan turned on\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "off" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0000;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Fan turned off\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else {
		printf ( "No such action\n" );
		exit ( 1 );
	};
}

static void
suspend ( char *instruction )
{
	reg_bx = 0x0010;

	if ( 0 == strcmp ( instruction, "status" )) {
		reg_ax = 0xfe00;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			if ( 0 == ( reg_cx & 0xffff )) {
				printf ( "APM is disabled (suspend is not possible)\n" );
			} else {
				printf ( "APM is enabled (suspend is possible)\n" );
			};
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "on" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0001;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Suspending...\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else {
		printf ( "No such action\n" );
		exit ( 1 );
	};
}

static void
display ()
{
	reg_bx = 0x0011;

	reg_ax = 0xfe00;
	hci ();
	if ( 0 == ( reg_ax & 0xff00 ) ) {
		printf ( "display type : " );
		switch ( reg_cx & 0xff ) {
			case 0: printf ( "STN monochrome" ); break;
			case 1: printf ( "STN color" ); break;
			case 2: printf ( "9 bit TFT" ); break;
			case 3: printf ( "12 bit TFT" ); break;
			case 4: printf ( "18 bit TFT" ); break;
			case 5: printf ( "24 bit TFT" ); break;
			default: printf ( "UNKNOWN" ); break;
		};
		printf ( ", resolution : " );
		switch ( ( reg_cx & 0xff00 ) / 256 ) {
			case 0: printf ( "640 x 480\n" ); break;
			case 1: printf ( "800 x 600\n" ); break;
			case 2: printf ( "1024 x 768\n" ); break;
			case 3: printf ( "1024 x 600\n" ); break;
			case 4: printf ( "800 x 480\n" ); break;
			default: printf ( "UNKNOWN\n" ); break;
		};
	} else {
		error ( ( reg_ax & 0xff00 ) / 256 );
	}
}

static void
selectbay ( char *instruction )
{
	reg_bx = 0x0014;
	reg_ax = 0xfe00;

	if ( 0 == strcmp ( instruction, "internal" )) {
		reg_cx = 0x0000;
	} else if ( 0 == strcmp ( instruction, "dock" )) {
		reg_cx = 0x0100;
	} else if ( 0 == strcmp ( instruction, "dockbay" )) {
		reg_cx = 0x0180;
	} else {
		printf ( "No such action\n" );
		exit ( 1 );
	};

	hci ();
	if ( 0 == ( reg_ax & 0xff00 ) ) {
		switch ( reg_cx ) {
			case 0: printf ( "Nothing inserted\n" ); break;
			case 1: printf ( "Floppy disk drive found\n" ); break;
			case 2: printf ( "ATAPI device found\n" ); break;
			case 3: printf ( "IDE device found\n" ); break;
			case 4: printf ( "Extra battery found\n" ); break;
			default: printf ( "UNKNOWN" ); break;
		};
	} else {
		error ( ( reg_ax & 0xff00 ) / 256 );
	}
}

static void
sysfifo ( char *instruction )
{
	reg_bx = 0x0016;

	if ( 0 == strcmp ( instruction, "status" )) {
		reg_ax = 0xfe00;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 )) {
			printf ( "System event %d\n", (int) ( reg_ax & 0xffff ));
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "on" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0001;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Successful\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "off" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0000;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Succcesful\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else {
		printf ( "No such action\n" );
		exit ( 1 );
	};
}

static void
panel ()
{
	reg_bx = 0x0019;

	reg_ax = 0xfe00;
	hci ();
	if ( 0 == ( reg_ax & 0xff00 ) ) {
		if ( 0 == ( reg_cx & 0xffff )) {
			printf ( "Panel closed\n" );
		} else {
			printf ( "Panel open\n" );
		};
	} else {
		error ( ( reg_ax & 0xff00 ) / 256 );
	};
}

static void
irda ( char *instruction )
{
	reg_bx = 0x001b;

	if ( 0 == strcmp ( instruction, "status" )) {
		reg_ax = 0xfe00;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			if ( 0 == ( reg_cx & 0xffff )) {
				printf ( "IrDA port is set to SIR\n" );
			} else {
				printf ( "IrDA port is set to FIR\n" );
			};
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "fir" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0001;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "IrDA port is set to FIR\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "sir" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0000;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "IrDA port is set to SIR\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else {
		printf ( "No such action\n" );
		exit ( 1 );
	};
}

static void
video ( char *instruction )
{
	reg_bx = 0x001c;

	if ( 0 == strcmp ( instruction, "status" )) {
		reg_ax = 0xfe00;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			if ( 0 == ( reg_cx & 0xffff )) {
				printf ( "Internal video output in use\n" );
			} else if ( 1 == ( reg_cx & 0xffff )) {
				printf ( "External video output in use\n" );
			} else {
				printf ( "Simultaneous video output in use\n" );
			};
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "internal" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0000;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Video output set to internal\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "external" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0001;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Video output set to external\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else if ( 0 == strcmp ( instruction, "simultaneous" )) {
		reg_ax = 0xff00;
		reg_cx = 0x0002;
		hci ();
		if ( 0 == ( reg_ax & 0xff00 ) ) {
			printf ( "Video output set to internal and external\n" );
		} else {
			error ( ( reg_ax & 0xff00 ) / 256 );
		}
	} else {
		printf ( "No such action\n" );
		exit ( 1 );
	};
}

static void
lock ( char *instruction )
{
	reg_bx = 0x0022;
	reg_ax = 0xfe00;

	if ( 0 == strcmp ( instruction, "builtin" )) {
		reg_cx = 0x0000;
	} else if ( 0 == strcmp ( instruction, "selectbay" )) {
		reg_cx = 0x0001;
	} else if ( 0 == strcmp ( instruction, "dock" )) {
		reg_cx = 0x0002;
	} else if ( 0 == strcmp ( instruction, "dockbay" )) {
		reg_cx = 0x0003;
	} else {
		printf ( "No such action\n" );
		exit ( 1 );
	};

	hci ();
	if ( 0 == ( reg_ax & 0xff00 ) ) {
		if ( 0 == ( reg_cx & 0xffff )) {
			printf ( "%s device is locked\n", instruction );
		} else {
			printf ( "%s device is not locked\n", instruction );
		};
	} else {
		error ( ( reg_ax & 0xff00 ) / 256 );
	}
}

int
main ( int argc, char **argv )
{
	if ( 3 != argc ) {
		if ( 2 == argc && 0 == strcmp ( argv[1], "help" )) {
			help ();
			exit ( 0 );
		} else if ( 2 == argc && 0 == strcmp ( argv[1], "display" )) {
			display ();
			exit ( 0 );
		} else if ( 2 == argc && 0 == strcmp ( argv[1], "power" )) {
			power ();
			exit ( 0 );
		} else if ( 2 == argc && 0 == strcmp ( argv[1], "panel" )) {
			panel ();
			exit ( 0 );
		} else {
			usage ( argv[0] );
		}
	};

	if ( 0 == strcmp( argv[1], "fan" )) {
		fan( argv[2] );
	} else if ( 0 == strcmp( argv[1], "backlight" )) {
		backlight( argv[2] );
	} else if ( 0 == strcmp( argv[1], "suspend" )) {
		suspend( argv[2] );
	} else if ( 0 == strcmp( argv[1], "selectbay" )) {
		selectbay( argv[2] );
	} else if ( 0 == strcmp( argv[1], "sysfifo" )) {
		sysfifo( argv[2] );
	} else if ( 0 == strcmp( argv[1], "irda" )) {
		irda( argv[2] );
	} else if ( 0 == strcmp( argv[1], "video" )) {
		video( argv[2] );
	} else if ( 0 == strcmp( argv[1], "lock" )) {
		lock( argv[2] );
 	} else {	
		usage( argv[0] );
	}

	return 0;
}
